// ScriptVersion 1.1
//
// Copyright (C) 2013 Red Cedar Technology - All Rights Reserved
//
// This script is used to update a STAR-CCM+ model for each design.
// Modifications to this script are allowed:
// - on a global basis by altering the script in the portals installation folder 
// - on a project-specific basis by adding the script as an input file to the
//   analysis and modifying the copy created in the project folder. If the script
//   is found in the analysis folder, it will take precedence over the script
//   in the portals installation folder.
// Modifications to this script should not remove the comments identifying where
// design-specific code is inserted.
//
// This script is not guaranteed to be backward compatible with previous or future
// versions of the HEEDS MDO STAR-CCM+ portal.

package macro;
import java.util.*;
import java.text.*;
import java.lang.reflect.Method;
import star.common.*;
import star.base.neo.*;
import star.base.report.*;
import star.energy.*;
import star.vis.*;
import star.flow.*;
import java.nio.charset.*;
import java.io.*;
import star.meshing.*;
import star.species.*;
import star.turbulence.*;
import star.cadmodeler.*;
import star.motion.*;
import star.starcad2.*;
// ###RCT_CAD_CLIENT_MODELS###
// this module only available for STAR-CCM+ 10.04 and above. 
// will be removed for older versions
import star.starcadbatch.common.*;
// end: RCT_CAD_CLIENT_MODELS

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.OutputKeys;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

public class updateSTARCCMModel extends StarMacro
{
	Simulation sim;
	Double clientVersion;
	
	@Override
	protected String resolveWorkPath() {
		return super.resolveWorkPath(); //To change body of generated methods, choose Tools | Templates.
  }
	
	// ###RCT_HEADER###  Do not remove this line - RCT inserts header code here
	String analysisName = "STARCCM_2";
	String modelFilename = "report_1_v3_new.sim"; 
	String version = "2020.01"; 
	int exportSceFile = 0; 
	String reports[] = { "Element Count All", "Element Count Bottom", "Element Count Top", "Element Count Volume", "Element Count planes1", "Eta", "Gamma", "Maximum MI_All", "Maximum MI_Bottom", "Maximum MI_Top", "Maximum MI_volume", "Minimum 1", "Minimum MI_All", "StartAvgConc", "Sum passive scalar All", "Sum passive scalar Bottom", "Sum passive scalar Top", "Sum passive scalar volume", "SumAvgConc", "SumConcSquared", "SumVolume", "Sum_kinetic_energy_times_volume", "Surface Standard Deviation 1", "Surface Standard Deviation All", "Surface Standard Deviation Bottom", "Surface Standard Deviation Top", "Total Solver CPU Time", "Volume Standard Deviation 1", "Volume Standard Deviation Volume", "Volume Uniformity Volume" };
	String reportResps[] = { "Element_Count_All_4", "Element_Count_Bottom_2", "Element_Count_Top_2", "Element_Count_Volume_2", "Element_Count_planes1_2", "Eta_2", "Gamma_val_2", "Maximum_MI_All_2", "Maximum_MI_Bottom_2", "Maximum_MI_Top_2", "Maximum_MI_volume_2", "Minimum_2", "Minimum_MI_All_2", "StartAvgConc_2", "Sum_passive_scalar_All_2", "Sum_passive_scalar_Bottom_2", "Sum_passive_scalar_Top_2", "Sum_passive_scalar_volume_2", "SumAvgConc_2", "SumConcSquared_2", "SumVolume_2", "Sum_kinetic_energy_times_volume_2", "Surface_Standard_Deviation_2", "Surface_Standard_Deviation_All_4", "Surface_Standard_Deviation_Bottom_4", "Surface_Standard_Deviation_Top_4", "Total_Solver_CPU_Time_4", "Volume_Standard_Deviation_4", "Volume_Standard_Deviation_Volume_2", "Volume_Uniformity_Volume_2" };
	String monitors[] = {};
	String monitorResps[] = {};
	String visualizationFiles[] = { "Residuals", "Maximum MI_All Monitor Plot", "Minimum MI_All Monitor Plot", "Maximum MI_Top Monitor Plot", "Maximum MI_Bottom Monitor Plot", "Element Count Bottom Monitor Plot", "Element Count All Monitor Plot", "Element Count Top Monitor Plot", "Sum passive scalar All Monitor Plot", "Sum passive scalar Bottom Monitor Plot", "Sum passive scalar Top Monitor Plot", "Surface Standard Deviation All Monitor Plot", "surface standar deviation 3 curves", "sum passive scalars 3 curves", "Monitors MI 3 curves max_csv", "Monitors MI 3 curves max jpg", "Sum_kinetic_energy_times_volume Monitor Plot", "Gamma Monitor Plot", "SumAvgConc Monitor Plot", "SumVolume Monitor Plot", "Maximum MI_volume Monitor Plot", "Five mixing indices", "Volume Standard Deviation Volume Monitor Plot", "Total Solver CPU Time Monitor Plot", "Mesh Scene z-axis_view", "Vector Scene 0.5 rad", "Scalar Scene  z-axis view1", "Vector Scene 10u", "Vector Scene 1.0 rad", "Geometry Scene two_section_planes", "Vector Scene 20u", "Vector Scene 30u", "Vector Scene 40u", "Vector Scene 60u", "Vector Scene 50u", "Vector Scene 90u", "Vector Scene 80u", "Vector Scene 70u", "Vector Scene 100u", "Vector Scene 420u", "Vector Scene 400u", "Vector Scene 430u", "Vector Scene 440u", "Vector Scene 410u", "Vector Scene 450u", "Vector Scene 460u", "Vector Scene 470u", "Vector Scene 480u", "Vector Scene 490u", "Scalar passive scalar with streamlines velocity all planes", "Passive scalar_2_planes", "Scalar Scene velocity stream lines magnitude", "Scalar Scene velocity stream lines z-direction", "Geometry Scene 1", "Vector Scene Outside", "Vector Scene 1.5 rad", "Vector Scene 2.0 rad", "Vector Scene 2.5 rad", "Vector Scene 3.0 rad", "Scalar Scene 0.5 rad", "Mesh Scene 1" };
	int vizFileWidth = 948;
	int vizFileHeight = 522;
	// end: RCT_HEADER
	
	public void execute()
	{ 
		clientVersion = Double.parseDouble(ClientVersion.getVersion().substring(0, ClientVersion.getVersion().lastIndexOf(".")));
		sim = getActiveSimulation();

		try {
			sim = getActiveSimulation();
			DateFormat dateFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
			// Do not remove the following code fences - RCT inserts update code here
			// ###RCT_GEOM###
			updatePart("model_cylinders_test-v2", "herringbone2.x_t");
			// end: RCT_GEOM

			// ###RCT_TAGS###
			int needOperation = 0;
			updateScalarGlobalParameter("basesize", "4e-5", needOperation);
			updateScalarGlobalParameter("bottom_velocity", "0.000800000", needOperation);
			updateScalarGlobalParameter("stoptime", "1", needOperation);
			updateScalarGlobalParameter("timestep", "10e-3", needOperation);
			updateScalarGlobalParameter("top_velocity", "0.000550000", needOperation);
			// end: RCT_TAG

			// ###RCT_CAD_CLIENT###
			// end: RCT_CAD_CLIENT

			// ###RCT_MACRO_PREMESH###
			// end: RCT_MACRO_PREMESH###

			// ###RCT_UPDATE_PARTS### (for when no mesh is performed)
			// end: RCT_UPDATE_PARTS

			// ###RCT_EXECUTE_OPERATIONS###
			executeOperations(sim, needOperation);
			// end: RCT_EXECUTE_OPERATIONS

			// ###RCT_CLEARSOLUTION###
			clearHistoryAndField(sim);
			// end: RCT_CLEARSOLUTION

			// ###RCT_MESH###
			meshSolution(sim);
			// end: RCT_MESH

			// ###RCT_REMOVEINVALIDCELLS###
			removeInvalidCells(sim);
			// end: RCT_REMOVEINVALIDCELLS

			// ###RCT_MACRO_POSTMESH###
			Date postmeshStart;
			Date postmeshEnd;
			postmeshStart = new Date();
			sim.println("Playing macro: remove_mesh_with_large_number_of_elements.java");
			sim.println("Started @ " + dateFormat.format(postmeshStart));
			new StarScript(sim, new File("remove_mesh_with_large_number_of_elements.java")).play();
			postmeshEnd = new Date();
			sim.println("Ended @ " + dateFormat.format(postmeshEnd));
			// end: RCT_MACRO_POSTMESH###

			// ###RCT_MIN_EVAL###
			// end: RCT_MIN_EVAL

			// ###RCT_RUN###
			runSolution(sim);
			// end: RCT_RUN

			// ###RCT_MACRO_POSTRUN###
			Date postrunStart;
			Date postrunEnd;
			postrunStart = new Date();
			sim.println("Playing macro: export_five_MI.java");
			sim.println("Started @ " + dateFormat.format(postrunStart));
			new StarScript(sim, new File("export_five_MI.java")).play();
			postrunEnd = new Date();
			sim.println("Ended @ " + dateFormat.format(postrunEnd));
			// end: RCT_MACRO_POSTRUN###

			writeOutputsToFile(sim, reports, reportResps, monitors, monitorResps, analysisName, modelFilename, version);

			// ###RCT_SAVE###
			turnOffAutoSave(sim);
			// end: RCT_SAVE

			createPlotsAndScenes(sim, visualizationFiles, exportSceFile);
			writeSuccessFile(sim);
		}
		catch (Exception e)
		{
			sim.println("ERROR: Error in executing updateSTARCCMModel.java");
			sim.println(e);
		}
		finally
		{
			writeCompletionFile(sim);
		}
	}

	private void updateCoordinateDesignParameter(CadModel cadModel, String varname, int component, Double varValue)
	{
		String varName = new String(varname.getBytes(), StandardCharsets.UTF_8);
		CoordinateDesignParameter param = (CoordinateDesignParameter) cadModel.getDesignParameterManager().getObject(varName);
		CadModelCoordinate cadModelCoord = param.getQuantity();
		double vector[] = cadModelCoord.getValue().toDoubleArray();
		vector[component] = varValue;
		cadModelCoord.setComponents(vector[0], vector[1], vector[2]);

		// move coord system if applicable
		try {
			LabCoordinateSystem labCoordSys = sim.getCoordinateSystemManager().getLabCoordinateSystem();
			CartesianCoordinateSystem cartCoordSys = ((CartesianCoordinateSystem) labCoordSys.getLocalCoordinateSystemManager().getObject(varName));
			Coordinate origin = cartCoordSys.getOrigin();
			double[] originArray = origin.getValue().toDoubleArray();
			originArray[component] = varValue;
			origin.setValue(new DoubleVector(originArray));
		}
		catch (Exception ex) 
		{
			sim.println("ERROR: Error in updating the coordinate of design parameter " + varName);
			sim.println(ex);
		}
		
		return;
	}

	private void updateMeshBaseSize(Simulation sim, MeshContinuum meshContinuum, Double varValue)
	{
		if (meshContinuum instanceof PartsMeshContinuum) {
			AutoMeshOperation meshOperation = ((AutoMeshOperation) sim.get(MeshOperationManager.class).getObject("Automated Mesh"));
			meshOperation.getDefaultValues().get(BaseSize.class).setValue(varValue);
		}
		else
			meshContinuum.getReferenceValues().get(BaseSize.class).setValue(varValue);

		return;
	}

	private void updateMotionRotation(String motionname, Double varValue)
	{
		String motionName = new String(motionname.getBytes(), StandardCharsets.UTF_8);
		RotatingMotion rotatingMotion = ((RotatingMotion) sim.get(MotionManager.class).getObject(motionName));
		Method getRotationRate = null;
		Method getRotationSpec = null;
		Method setValue = null;
		try {
			getRotationRate = rotatingMotion.getClass().getMethod("getRotationRate");
			getRotationSpec = rotatingMotion.getClass().getMethod("getRotationSpecification");
			setValue = getRotationRate.invoke(rotatingMotion).getClass().getMethod(("setValue"));
			setValue.invoke(getRotationRate.invoke(rotatingMotion), varValue);
		} catch (Exception e) {
			try {
				getRotationRate = getRotationSpec.invoke(rotatingMotion).getClass().getMethod("getRotationRate");
				setValue = getRotationRate.invoke(getRotationSpec.invoke(rotatingMotion)).getClass().getMethod(("setValue"));
				setValue.invoke(getRotationRate.invoke(getRotationSpec.invoke(rotatingMotion)), varValue);
			} catch (Exception ex)
			{
				throw new RuntimeException("ERROR: Failed to update rotation rate " + motionname, ex);
			}
		}
		
		return;
	}

	private void updateMotionRotationAndTranslation(String motionname, int component, Double varValue)
	{
		String motionName = new String(motionname.getBytes(), StandardCharsets.UTF_8);
		RotatingAndTranslatingMotion motion = ((RotatingAndTranslatingMotion) sim.get(MotionManager.class).getObject(motionName));
		if (component == -1) {
			motion.getRotationRate().setValue(varValue);
		}
		else {
			DoubleVector translationalVelocity = motion.getTranslationVelocity().getVector();
			translationalVelocity.setComponent(component, varValue);
			motion.getTranslationVelocity().setVector(translationalVelocity);
		}
		
		return;
	}

	private void updateParts()
	{
		Collection<GeometryPart> parts = sim.get(SimulationPartManager.class).getParts();
		for (GeometryPart part : parts) {
			sim.get(SimulationPartManager.class).updateParts(new NeoObjectVector(new Object[] {part}));
		}

		return;
	}

	// ###RCT_TRANSFORMS###
	// TRANSFORMS - only available for STAR-CCM+ 10.04 and above. The code below will be removed for older versions
	private void updateTransform_translate(String operationname, String transformName, int component, Double varValue)
	{
		String operationName = new String(operationname.getBytes(), StandardCharsets.UTF_8);
		TransformPartsOperation transformPartsOperation = 
			((TransformPartsOperation) sim.get(MeshOperationManager.class).getObject(operationName));
		TranslationControl translationControl = 
			((TranslationControl) transformPartsOperation.getTransforms().getObject(transformName));
		Coordinate coordinate =
			translationControl.getTranslationVector();
		DoubleVector vector = coordinate.getValue();
		vector.setComponent(component, varValue);
		Units units0 = coordinate.getUnits0();
		Units units1 = coordinate.getUnits1();
		Units units2 = coordinate.getUnits2();
		coordinate.setCoordinate(units0, units1, units2, vector);

		return;
	}

	private void updateTransform_rotate(String operationname, String transformName, double angle)
	{
		String operationName = new String(operationname.getBytes(), StandardCharsets.UTF_8);
		TransformPartsOperation transformPartsOperation = 
			((TransformPartsOperation) sim.get(MeshOperationManager.class).getObject(operationName));
		RotationControl rotationControl =
			((RotationControl) transformPartsOperation.getTransforms().getObject(transformName));
		rotationControl.getAngle().setValue(angle);

		return;
	}

	private void updateTransform_scale(String operationname, String transformName, int component, Double varValue)
	{
		String operationName = new String(operationname.getBytes(), StandardCharsets.UTF_8);
		TransformPartsOperation transformPartsOperation = 
			((TransformPartsOperation) sim.get(MeshOperationManager.class).getObject(operationName));
		ScalingControl scalingControl = 
			((ScalingControl) transformPartsOperation.getTransforms().getObject(transformName));
		DoubleVector vector = scalingControl.getScaleFactor().getVector();
		vector.setComponent(component, varValue);
		scalingControl.getScaleFactor().setVector(vector);

		return;
	}
	// end: RCT_TRANSFORMS

	private void updatePart(String hierarchyPart, String varValue)
	{
		String hierarchy = new String(hierarchyPart.getBytes(), StandardCharsets.UTF_8);
		String hierarchyParts[] = hierarchy.split("[\\.|]");
		CadPart cadPart = null;
		try {
			CompositePart compositePart = ((CompositePart) sim.get(SimulationPartManager.class).getPart(hierarchyParts[0]));
			sim.println("Found assembly part: " + compositePart.toString());
			for (int i = 1; i < hierarchyParts.length - 1; i++)
				compositePart = (CompositePart) compositePart.getChildParts().getPart((hierarchyParts[i]));
				cadPart = ((CadPart) compositePart.getChildParts().getPart(hierarchyParts[hierarchyParts.length - 1]));
			} 
			catch (Exception ex1) {
				// For non-assembly part
				try {
					cadPart = ((CadPart) sim.get(SimulationPartManager.class).getPart(hierarchy));
				} 
				catch (Exception ex2) {
					sim.println(ex1.toString());
					sim.println(ex2.toString());
				}
		}
		sim.println("Swapping geometry part " + cadPart.toString() + " for " + varValue);
		if (varValue.endsWith(".iges"))
			cadPart.reimportPart(varValue, 0.001, false);
		else
			cadPart.reimportPart(varValue);

		return;
	}

	private void updateScalarQuantityDesignParameter(CadModel cadModel, String varname, Double varValue)
	{
		String varName = new String(varname.getBytes(), StandardCharsets.UTF_8);
		ScalarQuantityDesignParameter param = (ScalarQuantityDesignParameter) cadModel.getDesignParameterManager().getObject(varName);
		ScalarPhysicalQuantityInput scalarInput = param.getQuantityInput();
		scalarInput.setDefinition(varValue + "");
		param.setQuantityInput(scalarInput);

		return;
		}

	private void updateUserDesignParameter(CadModel cadModel, String varname, Double varValue)
	{
		String varName = new String(varname.getBytes(), StandardCharsets.UTF_8);
		UserDesignParameter param = (UserDesignParameter) cadModel.getDesignParameterManager().getObject(varName);
		ScalarPhysicalQuantityInput scalarInput = param.getQuantityInput();
		scalarInput.setDefinition(varValue + "");
		param.setQuantityInput(scalarInput);
		
		return;
	}

	// ###RCT_CAD_CLIENT_MODELS###
	// RCT_CAD_CLIENT_MODELS - only available for STAR-CCM+ 10.04 and above. The code below will be removed for older versions
	private void updateDoubleCadDesignParameter(String starCadDocName, String partname, String varname, Double varValue)
	{
		if (clientVersion < 10.04) {
			sim.println("WARNING: Current version of STAR-CCM+ is " + clientVersion.toString() + ".\nCAD clients are only available in STAR-CCM+ 10.04 and above");
			return;
		}
		String cadDocName = new String(starCadDocName.getBytes(), StandardCharsets.UTF_8);
		String varName = new String(varname.getBytes(), StandardCharsets.UTF_8);
		String partName = new String(partname.getBytes(), StandardCharsets.UTF_8);
		StarCadDocument cadDoc = sim.get(StarCadDocumentManager.class).getObject(cadDocName);
		StarCadDesignParameterDouble varDouble = 
			(StarCadDesignParameterDouble) cadDoc.getStarCadDesignParameters().getParameter(partName + "\\" + varName);
		varDouble.setParamValue(varValue);

		return;
	}

	private void updateIntCadDesignParameter(String starCadDocName, String partname, String varname, int intValue)
	{
		if (clientVersion < 10.04) {
			sim.println("WARNING: Current version of STAR-CCM+ is " + clientVersion.toString() + ".\nCAD clients are only available in STAR-CCM+ 10.04 and above");
			return;
		}

		String cadDocName = new String(starCadDocName.getBytes(), StandardCharsets.UTF_8);
		String varName = new String(varname.getBytes(), StandardCharsets.UTF_8);
		String partName = new String(partname.getBytes(), StandardCharsets.UTF_8);
		StarCadDocument cadDoc = sim.get(StarCadDocumentManager.class).getObject(cadDocName);
		StarCadDesignParameterInt varInt = 
			(StarCadDesignParameterInt) cadDoc.getStarCadDesignParameters().getParameter(partName + "\\" + varName);
		varInt.setParamValue(intValue);

		return;
		}

	// CAD CLIENT MODELS
	private void updateCadClient(String cadDocName, String cadFileDir, int saveFlag)
	{
		Boolean exist = false;
		String starCadDocName = new String(cadDocName.getBytes(), StandardCharsets.UTF_8);
		
		try {
			Collection<StarCadDocument> cadDocs = sim.get(StarCadDocumentManager.class).getObjects();
			for (StarCadDocument cadDoc : cadDocs) {
				if (cadDoc.getModelName().compareToIgnoreCase(starCadDocName) != 0) {
					continue;
				}
				exist = true;
				try {
					File newCadFileDir = new File(cadFileDir + "/" + cadDoc.getModelName());
					if (newCadFileDir.exists()) {
						cadDoc.setCadFileDir(cadFileDir);
					} else {
						String filePath = cadDoc.getCadFileDir();
						File cadFilepath = new File(filePath + "/" + cadDoc.getModelName());
						if (cadFilepath.exists()) {
							sim.println("CAD client model \"" + cadDoc.getModelName() + "\" was not found in: " + cadFileDir);
							sim.println("Updating CAD client model from source location: " + cadDoc.getCadFileDir());
						}
						else {
							sim.println("CAD client model \"" + cadDoc.getModelName() + "\" was not found in: " + cadFileDir + "\n or in the source location: " + filePath);
							throw new RuntimeException("Failed to update CAD client model \"" + cadDoc.getModelName());
						}
					}
					// Update model
					if (saveFlag == 1) {
						cadDoc.updateModelAndSave();
					} else {
						cadDoc.updateModel();
					}
						
					// Break loop if update is completed
					if (!cadDoc.isUpdateFailed()) {
						String msg = "CAD client model \"" + cadDoc.getModelName() + "\" has been successfully updated";
						if (saveFlag == 1) {
							msg += "and saved";
						}
						sim.println(msg);
					} 
				} catch (Exception ex) {
					// When update failed
					if (cadDoc.isUpdateFailed()) {
						Method getErrorCode = null;
						try {
							getErrorCode = cadDoc.getClass().getMethod("getUpdateErrorCode");
						} catch (Exception e) {
							getErrorCode = cadDoc.getClass().getMethod("getModelUpdateErrorCode");
						}
						String errorCode = getErrorCode.invoke(cadDoc).toString();
						if (getErrorCode.invoke(cadDoc).toString() == "FAILED_TO_UPDATE") {
							sim.println("Terminating design as error");
							throw new RuntimeException("ERROR: Failed to update CAD client model \"" + cadDoc.getModelName() + "\".", ex);
						} else {
							throw new RuntimeException("ERROR: Failed to update CAD client model \"" + cadDoc.getModelName() + "\".\nError code: " + errorCode, ex);
						}
					}
					else
						throw ex;
				}
			} 
		} catch (Exception e) {
			if (!exist ) {
				sim.println("No CAD client model was defined in \"" + starCadDocName + "\"");
			}
			else {
				throw new RuntimeException("ERROR: Failed to update CAD client model", e);
			}
		}
		
		return;
	}
	// end: RCT_CAD_CLIENT_MODELS

	private void updateUserFieldFunction(String varname, String varValue)
	{
		String varName = new String(varname.getBytes(), StandardCharsets.UTF_8);
		UserFieldFunction userFieldFunction = ((UserFieldFunction) sim.getFieldFunctionManager().getObject(varName));
		userFieldFunction.setDefinition(varValue);

		return;
	}

	// ###RCT_GLOBAL_PARAMETERS###
	// RCT_GLOBAL_PARAMETERS - only available for STAR-CCM+ 11.02 and above. The code below will be removed for older versions
	private void updateScalarGlobalParameter(String varname, String varValue, int needOperation)
	{
		String varName = new String(varname.getBytes(), StandardCharsets.UTF_8);
		try {
			ScalarGlobalParameter scalarGP = (ScalarGlobalParameter) sim.get(GlobalParameterManager.class).getObject(varName);
			scalarGP.getQuantity().setValue(Double.parseDouble(varValue));
			if (scalarGP.inUse() && needOperation == 0) {
				Vector<ClientServerObject> gpRefs = new Vector<ClientServerObject>();
				// get GP dependencies
				 // in old API
				 if (clientVersion < 12.04) {
				  Vector<ClientServerObject> refs = scalarGP.getReferences();
					for (ClientServerObject ref : refs) {
						if (ref instanceof InlineFieldFunction) {
							InlineFieldFunction iff = (InlineFieldFunction) ref;
							gpRefs = iff.getReferences();
							break;
						}
					}
				} else {
					// for version staring from 12.04 
					gpRefs = scalarGP.getReferences();
				}
				
				for (ClientServerObject gpRef : gpRefs) {
					ClientServerObject gpDep = gpRef;
					boolean isOperation = false;
					int level = 0;
					while (isOperation) {
						gpDep = gpDep.getParent();
						String paramName = gpDep.getPresentationName();
						if (paramName.equalsIgnoreCase("Operations") || gpDep instanceof TransformPartsOperation) {
							isOperation = true;
							break;
						} else if (gpDep instanceof Simulation || level >= 5) {
							isOperation = false;
							break;
						}
						level++;
					}
					
					if (isOperation) {
						needOperation = 1;
						break;
					}
				}
			}
		} catch (Exception ex) {
			sim.println("ERROR: Error in updating scalar global parameter " + varName );
		}
		
		return;
	}

	private void updateVectorGlobalParameter(String varname, String varValue, int component, int needOperation) 
	{
		String varName = new String(varname.getBytes(), StandardCharsets.UTF_8);
		try {
			VectorGlobalParameter vectorGP = (VectorGlobalParameter) sim.get(GlobalParameterManager.class).getObject(varName);
			double vector[] = vectorGP.getQuantity().getVector().toDoubleArray();
			vector[component] = Double.parseDouble(varValue);
			vectorGP.getQuantity().setComponents(vector[0], vector[1], vector[2]);
			if (vectorGP.inUse() && needOperation == 0) {
				Vector<ClientServerObject> gpRefs = new Vector<ClientServerObject>();
				// get GP dependencies
				// in old API
				if (clientVersion < 12.04) {
					Vector<ClientServerObject> refs = vectorGP.getReferences();
					for (ClientServerObject ref : refs) {
						if (ref instanceof InlineFieldFunction) {
							InlineFieldFunction iff = (InlineFieldFunction) ref;
							gpRefs = iff.getReferences();
							break;
						}
					}
				} else {
					// for version staring from 12.04
					gpRefs = vectorGP.getReferences();
				}

				for (ClientServerObject gpRef : gpRefs) {
					ClientServerObject gpDep = gpRef;
					boolean isOperation = false;
					int level = 0;
					while (isOperation) {
						gpDep = gpDep.getParent();
						String paramName = gpDep.getPresentationName();
						if (paramName.equalsIgnoreCase("Operations") || gpDep instanceof TransformPartsOperation) {
							isOperation = true;
							break;
						} else if (gpDep instanceof Simulation || level >= 5) {
							isOperation = false;
							break;
						}
						level++;
					}
					
					if (isOperation) {
						needOperation = 1;
						break;
					}
				}
			}
		} catch (Exception ex) { 
			sim.println("ERROR: Error in updating vector global parameter " + varName );
		}

		return;
	}

	private void updateScalarGlobalParameter(String varName, String varValue)
	{
		int needOperation = 1;
		updateScalarGlobalParameter(varName, varValue, needOperation);
		
		return;
	}

	private void updateVectorGlobalParameter(String varName, String varValue, int component)
	{
		int needOperation = 1;
		updateVectorGlobalParameter(varName, varValue, component, needOperation);
		
		return;
	}
	// end: RCT_GLOBAL_PARAMETERS

	private void updateVectorQuantityDesignParameter(CadModel cadModel, String varname, int component, Double varValue)
	{
		String varName = new String(varname.getBytes(), StandardCharsets.UTF_8);
		VectorQuantityDesignParameter param = (VectorQuantityDesignParameter) cadModel.getDesignParameterManager().getObject(varName);
		double vector[] = param.getQuantity().getVector().toDoubleArray();
		vector[component] = varValue;
		param.getQuantity().setComponents(vector[0], vector[1], vector[2]);

		// move coord system if applicable
		try {
			LabCoordinateSystem labCoordSys = sim.getCoordinateSystemManager().getLabCoordinateSystem();
			CartesianCoordinateSystem cartCoordSys = ((CartesianCoordinateSystem) labCoordSys.getLocalCoordinateSystemManager().getObject(varName));
			Coordinate origin = cartCoordSys.getOrigin();
			double[] originArray = origin.getValue().toDoubleArray();
			originArray[component] = varValue;
			origin.setValue(new DoubleVector(originArray));
		} catch (Exception ex) {
			sim.println("ERROR: Error in updating vector quantity design parameter " + varName );
		}
		
		return;
	}

	// ###RCT_UPDATE_MONITORS###
	// RCT_UPDATE_MONITORS - only available for STAR-CCM+ 11.02 and above. The code below will be removed for older versions
	private void updateMonitorTriggerType(String monitors[], String triggers[])
	{
		try {
			for (int i = 0; i < monitors.length; ++i) {
				String monitorName = new String(monitors[i].getBytes(), StandardCharsets.UTF_8);
				String trigger = triggers[i];
				Monitor monitor = sim.getMonitorManager().getMonitor(monitorName);
				StarUpdate starUpdate = monitor.getStarUpdate();
				String triggerType = starUpdate.getUpdateModeInput().getSelected().toString();
				switch (triggerType) {
					case "ITERATION":
						triggerType = "0";
						break;
					case "TIMESTEP":
						triggerType = "1";
						break;
					case "DELTATIME":
						triggerType = "2";
						break;
					default:
						String warnMsg = "WARNING: " + triggerType + " is currently not supported. Switched to ITERATION.";
						sim.println(warnMsg);
						triggerType = "0";
				}
				if (triggerType == trigger) {
					continue;
				}
				switch (trigger) {
					case "0":
						starUpdate.getUpdateModeOption().setSelected(StarUpdateModeOption.Type.ITERATION);
						break;
					case "1":
						starUpdate.getUpdateModeOption().setSelected(StarUpdateModeOption.Type.TIMESTEP);
						break;
					case "2":
						starUpdate.getUpdateModeOption().setSelected(StarUpdateModeOption.Type.DELTATIME);
						break;
				}
			}
		} catch (Exception e) {
			throw new RuntimeException("ERROR: Error updating the monitor trigger.", e);
		}

		return;
	}
	// end: RCT_UPDATE_MONITORS

	private void executeOperations(Simulation sim)
	{
		int needOperation = 1;
		executeOperations(sim, needOperation);

		return;
	}

	private void executeOperations(Simulation sim, int needOperation)
	{
		if(needOperation == 1)
			sim.get(MeshOperationManager.class).executeAll();
		
		return;
	}

	private void clearHistoryOnly(Simulation sim)
	{
		sim.getSolution().clearSolution(Solution.Clear.History);

		return;
	}

	private void clearHistoryAndField(Simulation sim)
	{
		sim.getSolution().clearSolution();
		return;
	}

	private void meshSolution(Simulation sim)
	{
		MeshPipelineController meshPipelineController = sim.get(MeshPipelineController.class);
		meshPipelineController.generateVolumeMesh();

		return;
	}

	private void removeInvalidCells(Simulation sim)
	{
		MeshManager meshManager = sim.getMeshManager();
		meshManager.removeInvalidCells(new NeoObjectVector(sim.getRegionManager().getRegions().toArray()), 0.51, 1.0E-8, 1.0E-10, 1, 0.0, 0.0);

		return;
	}

	private void runMinIteration(int nStep)
	{
		sim.getSimulationIterator().step(nStep);
		return;
	}

	private void runSolution(Simulation sim)
	{
		sim.getSimulationIterator().run();

		return;
	}

	private Element createElementWithAttributes(Document doc, String elemName, String attributeList[], String attributeValueList[], double textValue)
	{
		Element child = doc.createElement(elemName);
		for (int i=0; i<attributeList.length; i++) {
			child.setAttribute(attributeList[i], attributeValueList[i]);
		}
		String valAsString = String.valueOf(textValue);
		if (valAsString.length() > 0) {
			child.appendChild(doc.createTextNode(valAsString));
		}

		return child;
	}

	private void writeOutputsToFile(Simulation sim, String reports[], String responseNames[], String analysisName, String modelFilename, String version)
	{
		try
		{
			// Create root node of xml doc
			DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder = docFactory.newDocumentBuilder();
			Document myXmlDoc = docBuilder.newDocument();
			Element root = myXmlDoc.createElement("Filename");
			String modelFileNode = new String(modelFilename.getBytes(), StandardCharsets.UTF_8);
			root.setAttribute("name", modelFileNode);
			root.setAttribute("solver", "STAR-CCM+");
			root.setAttribute("version", version);
			myXmlDoc.appendChild(root);

			// Create child elements for each report
			for (int i=0; i<reports.length; i++)
			{
				String reportName = new String(reports[i].getBytes(), StandardCharsets.UTF_8);
				String responseName = new String(responseNames[i].getBytes(), StandardCharsets.UTF_8);
				Report report = sim.getReportManager().getReport(reportName);
				double reportValue = report.getReportMonitorValue();
				String[] attNames = {"name", "output"};
				String[] attValues = {responseName, reportName};
				Element responseNode = createElementWithAttributes(myXmlDoc, "Response", attNames, attValues, reportValue);
				root.appendChild(responseNode);
			}

			// Write xml doc
			String modelFilenameNoExt = modelFileNode.replaceFirst("[.][^.]+$", "");
			String anlName = new String(analysisName.getBytes(), StandardCharsets.UTF_8);
			String xmlfile = "results_" + analysisName + "_" + modelFilenameNoExt + ".xml";
			Transformer transformer = TransformerFactory.newInstance().newTransformer();
			transformer.setOutputProperty(OutputKeys.INDENT, "yes");
			transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
			DOMSource source = new DOMSource(myXmlDoc);
			StreamResult result = new StreamResult(new File(xmlfile));
			transformer.transform(source, result);
		}
		catch (Exception ex)
		{
			sim.println(ex);
			throw new RuntimeException("ERROR: Error in writing output file.", ex);
		}
	}

	private void writeOutputsToFile(Simulation sim, String reports[], String reportResps[], String monitors[], String monitorResps[], String analysisName, String modelFilename, String version)
	{
		writeOutputsToFile(sim, reports, reportResps, analysisName, modelFilename, version);
		if (monitors.length < 1)
			return;
		
		// Export monitors
		String cwd = resolveWorkPath() + "/";
		sim.println("Looking .csv files in " + cwd + "...\n");

		for (int j = 0; j < monitors.length; j++) {
			String monitorName = new String(monitors[j].getBytes(), StandardCharsets.UTF_8);
			String respsName = new String(monitorResps[j].getBytes(), StandardCharsets.UTF_8);

			try {
				Object monitorObj = sim.getMonitorManager().getMonitor(monitorName);

				if (monitorObj instanceof ReportMonitor) {
					ReportMonitor monitor = (ReportMonitor) monitorObj;
					sim.getMonitorManager().export(cwd + respsName + ".csv", ",", new NeoObjectVector(new Object[]{monitor}));

				} else if (monitorObj instanceof ResidualMonitor) {
					ResidualMonitor monitor = (ResidualMonitor) monitorObj;
					sim.getMonitorManager().export(cwd + respsName + ".csv", ",", new NeoObjectVector(new Object[]{monitor}));

				} else if (monitorObj instanceof SimulationIteratorTimeReportMonitor) {
					SimulationIteratorTimeReportMonitor monitor = (SimulationIteratorTimeReportMonitor) monitorObj;
					sim.getMonitorManager().export(cwd + respsName + ".csv", ",", new NeoObjectVector(new Object[]{monitor}));
				} else {
					sim.println("Selected monitor \"" + monitorName + "\" is currently not supported.");
					continue;
				}
			} catch (Exception ex) {
				throw new RuntimeException("ERROR: Error in exporting CSV file for monitor response " + respsName, ex);
			}
		}

		return;

	}

	private void saveSim(Simulation sim)
	{
		sim.saveState(resolvePath(sim.getPresentationName()) + ".sim");

		return;
	}

	private void turnOffAutoSave(Simulation sim)
	{
		AutoSave autoSave = sim.getSimulationIterator().getAutoSave();
		autoSave.setAutoSaveBatch(false);

		return;
	}

	private void createPlotsAndScenes(Simulation sim, String visualizationFiles[], int exportSce)
	{
		if (visualizationFiles.length > 0)
		{
			sim.println("Starting to create visualization files:");
			for (String s : visualizationFiles)
			{
				String fname;
				String vizname = new String(s.getBytes(), StandardCharsets.UTF_8);
				try {
					StarPlot plot = sim.getPlotManager().getPlot(vizname);
					fname = plot.getPresentationName() + ".png";
					sim.println("\t" + fname);
					plot.encode(fname, vizFileWidth, vizFileHeight);
				}
				catch (Exception ex)
				{
					try {
						Scene scene = sim.getSceneManager().getScene(vizname);
						fname = scene.getPresentationName() + ".png";
						sim.println("\t" + fname);
						scene.printAndWait(fname, 1, vizFileWidth, vizFileHeight); // save png
						if (exportSce == 1) {
							fname = scene.getPresentationName() + ".sce";
							scene.export3DSceneFileAndWait(fname, "CAD-" + scene.getPresentationName(), "", true, true);
						}
					}
					catch (Exception e)
					{
						sim.println("ERROR: Error in creating visualization files.");
						sim.println(e);
					}
				}
			}
			sim.println("Finished creating visualization files.");
		}

		return;
	}

	private void createPlotsAndScenes(Simulation sim, String visualizationFiles[])
	{
		int exportSce = 0;
		createPlotsAndScenes(sim, visualizationFiles, exportSce);
	}

	private void writeCompletionFile(Simulation sim)
	{
		try {
			File outFile = new File(sim.getSessionDir() + System.getProperty("file.separator") + "DONE");
			outFile.createNewFile();
		}
		catch(Exception ex)
		{
			sim.println("ERROR: Error in writing completion file.");
			sim.println(ex);
		}

		return;
	}

	private void writeSuccessFile(Simulation sim)
	{
		try {
			File outFile = new File(sim.getSessionDir() + System.getProperty("file.separator") + "SUCCESS");
			outFile.createNewFile();
		}
		catch(Exception ex)
		{
			sim.println("ERROR: Error in writing success file.");
			sim.println(ex);
		}

		return;
	}
	// ###RCT_FUNCTIONS### Do not remove this line - RCT inserts extra functions here
	// end:: RCT_FUNCTIONS
}


